package gregapi.recipes.handlers;

import static gregapi.data.CS.F;
import static gregapi.data.CS.T;
import static gregapi.data.CS.U;
import gregapi.code.ICondition;
import gregapi.data.TD;
import gregapi.oredict.OreDictItemData;
import gregapi.oredict.OreDictMaterial;
import gregapi.oredict.OreDictPrefix;
import gregapi.recipes.IRecipeMapHandler;
import gregapi.recipes.Recipe.RecipeMap;
import gregapi.util.UT;
import net.minecraft.item.ItemStack;
import net.minecraftforge.fluids.Fluid;
import net.minecraftforge.fluids.FluidStack;

/**
 * @author Gregorius Techneticies
 */
public class RecipeMapHandlerPrefix implements IRecipeMapHandler {
	private final ICondition<OreDictMaterial> mCondition;
	private final OreDictPrefix[] mInputPrefixes, mOutputPrefixes;
	private final byte[] mInputAmounts, mOutputAmounts;
	private final ItemStack mAdditionalInput;
	private final FluidStack mFluidInputPerUnit, mFluidOutputPerUnit;
	private final long mUnitsProcessed, mEUt, mDuration, mMultiplier;
	private final boolean mAllowToGenerateAllRecipesAtOnce;
	
	private boolean mAlreadyAddedAllRecipes = F;
	
	public RecipeMapHandlerPrefix(OreDictPrefix aInputPrefix, long aInputAmount, FluidStack aFluidInputPerUnit, long aEUt, long aDuration, long aMultiplier, FluidStack aFluidOutputPerUnit, OreDictPrefix aOutputPrefix, long aOutputAmount, ItemStack aAdditionalInput, boolean aAllowToGenerateAllRecipesAtOnce, ICondition<OreDictMaterial> aCondition) {
		mAllowToGenerateAllRecipesAtOnce = aAllowToGenerateAllRecipesAtOnce;
		mFluidInputPerUnit	= aFluidInputPerUnit;
		mFluidOutputPerUnit	= aFluidOutputPerUnit;
		mCondition			= aCondition;
		mInputPrefixes		= new OreDictPrefix[] {aInputPrefix};
		mInputAmounts		= new byte[] {UT.Code.bindStack(aInputAmount)};
		mOutputPrefixes		= new OreDictPrefix[] {aOutputPrefix};
		mOutputAmounts		= new byte[] {UT.Code.bindStack(aOutputAmount)};
		mAdditionalInput	= aAdditionalInput;
		long tUnitsProcessed = 0;
		for (int i = 0; i < mInputPrefixes.length; i++) tUnitsProcessed += mInputPrefixes[i].mAmount * mInputAmounts[i];
		mUnitsProcessed		= tUnitsProcessed;
		mMultiplier			= aMultiplier;
		mDuration 			= aDuration;
		mEUt				= aEUt;
	}
	
	public RecipeMapHandlerPrefix(OreDictPrefix aInputPrefix1, long aInputAmount1, OreDictPrefix aInputPrefix2, long aInputAmount2, FluidStack aFluidInputPerUnit, long aEUt, long aDuration, long aMultiplier, FluidStack aFluidOutputPerUnit, OreDictPrefix aOutputPrefix1, long aOutputAmount1, OreDictPrefix aOutputPrefix2, long aOutputAmount2, ItemStack aAdditionalInput, boolean aAllowToGenerateAllRecipesAtOnce, ICondition<OreDictMaterial> aCondition) {
		mAllowToGenerateAllRecipesAtOnce = aAllowToGenerateAllRecipesAtOnce;
		mFluidInputPerUnit	= aFluidInputPerUnit;
		mFluidOutputPerUnit	= aFluidOutputPerUnit;
		mCondition			= aCondition;
		mInputPrefixes		= aInputPrefix2 == null ? new OreDictPrefix[] {aInputPrefix1} : new OreDictPrefix[] {aInputPrefix1, aInputPrefix2};
		mInputAmounts		= aInputPrefix2 == null ? new byte[] {UT.Code.bindStack(aInputAmount1)} : new byte[] {UT.Code.bindStack(aInputAmount1), UT.Code.bindStack(aInputAmount2)};
		mOutputPrefixes		= aOutputPrefix2 == null ? new OreDictPrefix[] {aOutputPrefix1} : new OreDictPrefix[] {aOutputPrefix1, aOutputPrefix2};
		mOutputAmounts		= aOutputPrefix2 == null ? new byte[] {UT.Code.bindStack(aOutputAmount1)} : new byte[] {UT.Code.bindStack(aOutputAmount1), UT.Code.bindStack(aOutputAmount2)};
		mAdditionalInput	= aAdditionalInput;
		long tUnitsProcessed = 0;
		for (int i = 0; i < mInputPrefixes.length; i++) tUnitsProcessed += mInputPrefixes[i].mAmount * mInputAmounts[i];
		mUnitsProcessed		= tUnitsProcessed;
		mMultiplier			= aMultiplier;
		mDuration 			= aDuration;
		mEUt				= aEUt;
	}
	
	public RecipeMapHandlerPrefix(OreDictPrefix[] aInputPrefixes, long[] aInputAmount, FluidStack aFluidInputPerUnit, long aEUt, long aDuration, long aMultiplier, FluidStack aFluidOutputPerUnit, OreDictPrefix[] aOutputPrefixes, long[] aOutputAmount, ItemStack aAdditionalInput, boolean aAllowToGenerateAllRecipesAtOnce, ICondition<OreDictMaterial> aCondition) {
		mAllowToGenerateAllRecipesAtOnce = aAllowToGenerateAllRecipesAtOnce;
		mFluidInputPerUnit	= aFluidInputPerUnit;
		mFluidOutputPerUnit	= aFluidOutputPerUnit;
		mCondition			= aCondition;
		mInputPrefixes		= aInputPrefixes;
		mInputAmounts		= new byte[mInputPrefixes.length];
		for (int i = 0; i < mInputAmounts.length; i++) mInputAmounts[i] = UT.Code.bindStack(aInputAmount[i]);
		mOutputPrefixes		= aOutputPrefixes;
		mOutputAmounts		= new byte[mOutputPrefixes.length];
		for (int i = 0; i < mOutputAmounts.length; i++) mOutputAmounts[i] = UT.Code.bindStack(aOutputAmount[i]);
		mAdditionalInput	= aAdditionalInput;
		long tUnitsProcessed = 0;
		for (int i = 0; i < mInputPrefixes.length; i++) tUnitsProcessed += mInputPrefixes[i].mAmount * mInputAmounts[i];
		mUnitsProcessed		= tUnitsProcessed;
		mMultiplier			= aMultiplier;
		mDuration 			= aDuration;
		mEUt				= aEUt;
	}
	
	@Override
	public boolean addRecipesUsing(RecipeMap aMap, ItemStack aStack, OreDictItemData aData) {
		if (isDone()) return F;
		if (UT.Stacks.equal(aStack, mAdditionalInput)) return mAllowToGenerateAllRecipesAtOnce && addAllRecipesInternal(aMap);
		return aData != null && aData.hasValidMaterialData() && UT.Code.contains(aData.mPrefix, mInputPrefixes) && addRecipeForMaterial(aMap, aData.mMaterial.mMaterial);
	}
	
	@Override
	public boolean addRecipesProducing(RecipeMap aMap, ItemStack aStack, OreDictItemData aData) {
		if (isDone()) return F;
		return aData != null && aData.hasValidMaterialData() && UT.Code.contains(aData.mPrefix, mOutputPrefixes) && addRecipeForMaterial(aMap, aData.mMaterial.mMaterial);
	}
	
	@Override
	public boolean containsInput(RecipeMap aMap, ItemStack aStack, OreDictItemData aData) {
		if (isDone()) return F;
		return addRecipesUsing(aMap, aStack, aData);
	}
	
	@Override
	public boolean addRecipesUsing(RecipeMap aMap, Fluid aFluid) {
		if (isDone()) return F;
		return mFluidInputPerUnit != null && mFluidInputPerUnit.getFluid() == aFluid && mAllowToGenerateAllRecipesAtOnce && addAllRecipesInternal(aMap);
	}
	
	@Override
	public boolean addRecipesProducing(RecipeMap aMap, Fluid aFluid) {
		if (isDone()) return F;
		return mFluidOutputPerUnit != null && mFluidOutputPerUnit.getFluid() == aFluid && mAllowToGenerateAllRecipesAtOnce && addAllRecipesInternal(aMap);
	}
	
	@Override
	public boolean containsInput(RecipeMap aMap, Fluid aFluid) {
		return mFluidInputPerUnit != null && mFluidInputPerUnit.getFluid() == aFluid;
	}
	
	@Override
	public boolean addAllRecipes(RecipeMap aMap) {
		return mAllowToGenerateAllRecipesAtOnce && addAllRecipesInternal(aMap);
	}
	
	public boolean addAllRecipesInternal(RecipeMap aMap) {
		if (isDone()) return F;
		for (OreDictMaterial tMaterial : mInputPrefixes[0].mRegisteredMaterials) addRecipeForMaterial(aMap, tMaterial);
		mAlreadyAddedAllRecipes = T;
		return T;
	}
	
	@Override
	public boolean isDone() {
		return mAlreadyAddedAllRecipes;
	}
	
	public boolean addRecipeForMaterial(RecipeMap aMap, OreDictMaterial aMaterial) {
		if (!mCondition.isTrue(aMaterial) || aMaterial.contains(TD.Properties.INVALID_MATERIAL)) return F;
		
		ItemStack[] tInputs = new ItemStack[mAdditionalInput==null?mInputPrefixes.length:mInputPrefixes.length+1];
		if (mAdditionalInput != null) tInputs[tInputs.length-1] = mAdditionalInput;
		for (int i = 0; i < mInputPrefixes.length; i++) tInputs[i] = mInputPrefixes[i].mat(aMaterial, mInputAmounts[i]);
		for (ItemStack tInput : tInputs) if (UT.Stacks.invalid(tInput)) return F;
		
		ItemStack[] tOutputs = new ItemStack[mOutputPrefixes.length];
		OreDictMaterial tMaterial = getOutputMaterial(aMaterial);
		for (int i = 0; i < mOutputPrefixes.length; i++) tOutputs[i] = mOutputPrefixes[i].mat(tMaterial, mOutputAmounts[i]);
		for (ItemStack tOutput : tOutputs) if (UT.Stacks.invalid(tOutput)) return F;
		
		return aMap.addRecipeX(F,T,F,F,T, mEUt, mDuration<=0?getCosts(aMaterial):mDuration, tInputs, UT.Fluids.mul(mFluidInputPerUnit, aMaterial.mToolQuality+1), UT.Fluids.mul(mFluidOutputPerUnit, aMaterial.mToolQuality+1), tOutputs) != null;
	}
	
	public OreDictMaterial getOutputMaterial(OreDictMaterial aMaterial) {
		return aMaterial;
	}
	
	public long getCosts(OreDictMaterial aMaterial) {
		return UT.Code.units(mUnitsProcessed, U, mMultiplier+mMultiplier*aMaterial.mToolQuality, T);
	}
}