package gregapi.worldgen;

import static gregapi.data.CS.*;
import gregapi.block.IBlockPlacable;
import gregapi.code.ArrayListNoNulls;
import gregapi.code.ItemStackContainer;
import gregapi.data.CS.BlocksGT;
import gregapi.oredict.OreDictMaterial;

import java.util.List;
import java.util.Random;

import net.minecraft.block.Block;
import net.minecraft.init.Blocks;
import net.minecraft.world.World;
import net.minecraft.world.biome.BiomeGenBase;
import net.minecraft.world.chunk.Chunk;
import net.minecraft.world.chunk.IChunkProvider;
import net.minecraft.world.gen.ChunkProviderEnd;
import net.minecraft.world.gen.ChunkProviderHell;
import net.minecraftforge.common.util.ForgeDirection;

/**
 * @author Gregorius Techneticies
 */
public class Worldgenerator implements cpw.mods.fml.common.IWorldGenerator {
	public static boolean sAsteroids = T;
	
	public List<Runnable> mList = new ArrayListNoNulls<Runnable>();
	public static boolean sIsGenerating = F, sCanGenerateUniqueFeatureForThisChunk = T;
	
	/** Lists of all the active World generation Features, these are getting initialised in Load! */
	public static final List<WorldgenObject> sWorldgenList = new ArrayListNoNulls<WorldgenObject>();
	
	public Worldgenerator() {
		cpw.mods.fml.common.registry.GameRegistry.registerWorldGenerator(this, Integer.MAX_VALUE / 2);
	}
	
	@Override
    public void generate(Random aRandom, int aX, int aZ, World aWorld, IChunkProvider aChunkGenerator, IChunkProvider aChunkProvider) {
		mList.add(new WorldGenContainer(new Random(aRandom.nextInt()), aX * 16, aZ * 16, aWorld.provider.dimensionId == -1 || aChunkGenerator instanceof ChunkProviderHell || aWorld.getBiomeGenForCoords(aX*16+8, aZ*16+8) == BiomeGenBase.hell ? -1 : aWorld.provider.dimensionId == +1 || aChunkGenerator instanceof ChunkProviderEnd || aWorld.getBiomeGenForCoords(aX*16+8, aZ*16+8) == BiomeGenBase.sky ? +1 : 0, aWorld, aChunkGenerator, aChunkProvider, aWorld.getBiomeGenForCoords(aX*16+8, aZ*16+8).biomeName));
		if (!sIsGenerating) {
			sIsGenerating = T;
			while (!mList.isEmpty()) mList.remove(0).run();
			sIsGenerating = F;
		}
	}
	
	public static class WorldGenContainer implements Runnable {
		public final Random mRandom;
		public final int mX, mZ, mDimensionType;
		public final World mWorld;
		public final IChunkProvider mChunkGenerator, mChunkProvider;
		public final String mBiome;
		
		public WorldGenContainer(Random aRandom, int aX, int aZ, int aDimensionType, World aWorld, IChunkProvider aChunkGenerator, IChunkProvider aChunkProvider, String aBiome) {
			mRandom = aRandom;
			mX = aX;
			mZ = aZ;
			mDimensionType = aDimensionType;
			mWorld = aWorld;
			mChunkGenerator = aChunkGenerator;
			mChunkProvider = aChunkProvider;
			mBiome = aBiome;
		}
		
		@Override
		public void run() {
			for (int tX=-32; tX<=32; tX+=16) for (int tZ=-32; tZ<=32; tZ+=16) {
				int tChunkX = mX+tX;
				int tChunkZ = mZ+tZ;
		    	if (((tChunkX/16)+402653184) % 3 == 1 && ((tChunkZ/16)+402653184) % 3 == 1) {
		    		Random aRandom = new Random(mWorld.getSeed());
		    		aRandom = new Random(((aRandom.nextLong() >> 2 + 1L) * (tChunkX/16) + (aRandom.nextLong() >> 2 + 1L) * (tChunkZ/16)) ^ mWorld.getSeed());
		            
		    		if (Worldgen_GT_Ore_Layer.sWeight > 0 && Worldgen_GT_Ore_Layer.sList.size() > 0) {
		    			boolean temp = T;
		            	for (int i = 0; i < 256 && temp; i++) {
		                	int tRandomWeight = aRandom.nextInt(Worldgen_GT_Ore_Layer.sWeight);
		                	for (Worldgen_GT_Ore_Layer tWorldGen : Worldgen_GT_Ore_Layer.sList) {
		                		tRandomWeight -= tWorldGen.mWeight;
		                		if (tRandomWeight <= 0) {
		                    		try {
		                    			if (tWorldGen.executeWorldgen(mWorld, aRandom, mBiome, mDimensionType, mX, mZ, tChunkX, tChunkZ, mChunkGenerator, mChunkProvider)) {
		                    				temp = F;
		                    			}
		                    		} catch (Throwable e) {
		                    			e.printStackTrace(ERR);
		                    		}
                    				break;
		                		}
		                	}
		            	}
		    		}
		    	}
			}
	    	
    	    String tBiome = mWorld.getBiomeGenForCoords(mX+8, mZ+8).biomeName;
    	    if (tBiome == null) tBiome = BiomeGenBase.plains.biomeName;
    	    sCanGenerateUniqueFeatureForThisChunk = T;
            for (WorldgenObject tWorldGen : sWorldgenList) {
            	try {
            		tWorldGen.executeWorldgen(mWorld, mRandom, mBiome, mDimensionType, mX, mZ, mChunkGenerator, mChunkProvider);
            	} catch (Throwable e) {
            		e.printStackTrace(ERR);
            	}
    		}
    	    sCanGenerateUniqueFeatureForThisChunk = T;
	    	
			Chunk tChunk = mWorld.getChunkFromBlockCoords(mX, mZ);
	    	if (tChunk != null) tChunk.isModified = T;
		}
	}
	
	public static boolean isReplaceable(World aWorld, int aX, int aY, int aZ, boolean aAllowAir) {
		Block aBlock = aWorld.getBlock(aX, aY, aZ);
		if (aBlock == NB) return aAllowAir;
		byte aMetaData = (byte)aWorld.getBlockMetadata(aX, aY, aZ);
		if (BlocksGT.stoneToNormalOres.containsKey(new ItemStackContainer(aBlock, 1, aMetaData))) return T;
		if (aBlock.isReplaceableOreGen(aWorld, aX, aY, aZ, Blocks.stone				)) return T;
		if (aBlock.isReplaceableOreGen(aWorld, aX, aY, aZ, Blocks.gravel			)) return T;
		if (aBlock.isReplaceableOreGen(aWorld, aX, aY, aZ, Blocks.sand				)) return T;
		if (aBlock.isReplaceableOreGen(aWorld, aX, aY, aZ, Blocks.netherrack		)) return T;
		if (aBlock.isReplaceableOreGen(aWorld, aX, aY, aZ, Blocks.end_stone			)) return T;
		return F;
	}
	
	public static boolean setOreBlock(World aWorld, int aX, int aY, int aZ, OreDictMaterial aMaterial) {
		return aMaterial != null && setOreBlock(aWorld, aX, aY, aZ, aMaterial.mID);
	}
	
	public static boolean setOreBlock(World aWorld, int aX, int aY, int aZ, short aID) {
		if (aID <= 0 && aID == W) return false;
		Block aBlock = aWorld.getBlock(aX, aY, aZ);
		IBlockPlacable tBlock = BlocksGT.stoneToNormalOres.get(new ItemStackContainer(aBlock, 1, aWorld.getBlockMetadata(aX, aY, aZ)));
		if (tBlock == null) {
		if (aBlock.isReplaceableOreGen(aWorld, aX, aY, aZ, Blocks.stone				)) tBlock = BlocksGT.ore; else
		if (aBlock.isReplaceableOreGen(aWorld, aX, aY, aZ, Blocks.gravel			)) tBlock = BlocksGT.oreGravel; else
		if (aBlock.isReplaceableOreGen(aWorld, aX, aY, aZ, Blocks.sand				)) tBlock = BlocksGT.oreSand; else
		if (aBlock.isReplaceableOreGen(aWorld, aX, aY, aZ, Blocks.netherrack		)) tBlock = BlocksGT.oreNetherrack; else
		if (aBlock.isReplaceableOreGen(aWorld, aX, aY, aZ, Blocks.end_stone			)) tBlock = BlocksGT.oreEndstone;
		}
		return tBlock != null && tBlock.placeBlock(aWorld, aX, aY, aZ, (byte)6, aID, null, F, T);
	}
	
	public static boolean setSmallOreBlock(World aWorld, int aX, int aY, int aZ, OreDictMaterial aMaterial) {
		return aMaterial != null && setSmallOreBlock(aWorld, aX, aY, aZ, aMaterial.mID);
	}
	
	public static boolean setSmallOreBlock(World aWorld, int aX, int aY, int aZ, short aID) {
		if (aID <= 0 && aID == W) return false;
		Block aBlock = aWorld.getBlock(aX, aY, aZ);
		IBlockPlacable tBlock = BlocksGT.stoneToSmallOres.get(new ItemStackContainer(aBlock, 1, aWorld.getBlockMetadata(aX, aY, aZ)));
		if (tBlock == null) {
		if (aBlock.isReplaceableOreGen(aWorld, aX, aY, aZ, Blocks.stone				)) tBlock = BlocksGT.oreSmall; else
		if (aBlock.isReplaceableOreGen(aWorld, aX, aY, aZ, Blocks.gravel			)) tBlock = BlocksGT.oreSmallGravel; else
		if (aBlock.isReplaceableOreGen(aWorld, aX, aY, aZ, Blocks.sand				)) tBlock = BlocksGT.oreSmallSand; else
		if (aBlock.isReplaceableOreGen(aWorld, aX, aY, aZ, Blocks.netherrack		)) tBlock = BlocksGT.oreSmallNetherrack; else
		if (aBlock.isReplaceableOreGen(aWorld, aX, aY, aZ, Blocks.end_stone			)) tBlock = BlocksGT.oreSmallEndstone;
		}
		return tBlock != null && tBlock.placeBlock(aWorld, aX, aY, aZ, (byte)6, aID, null, F, T);
	}
	
	/** Removes Bedrock from that Position and replaces it with regular Stone of the region. */
	public static boolean removeBedrock(World aWorld, int aX, int aY, int aZ) {
		if (aWorld.getBlock(aX, aY, aZ) == Blocks.bedrock) for (int i = 1; i < 10; i++) for (ForgeDirection tDirection : ForgeDirection.VALID_DIRECTIONS) {
			Block tBlock = aWorld.getBlock(aX+tDirection.offsetX*i, aY+tDirection.offsetY*i, aZ+tDirection.offsetZ*i);
			if (tBlock != Blocks.bedrock && tBlock != NB) {
				int tMetaData = aWorld.getBlockMetadata(aX+tDirection.offsetX*i, aY+tDirection.offsetY*i, aZ+tDirection.offsetZ*i);
				if (BlocksGT.stoneToNormalOres.containsKey(new ItemStackContainer(tBlock, 1, tMetaData))) {
					return aWorld.setBlock(aX, aY, aZ, tBlock, tMetaData, 0);
				}
			}
		}
		return aWorld.setBlock(aX, aY, aZ, Blocks.stone, 0, 0);
	}
}